/*
 *  Copyright (C) 2015 Jeroen Domburg <jeroen at spritesmods.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "voom.h"
#include "math.h"
#include <stdlib.h>
#include <stdio.h>
#ifndef VECX
#include "fatfs/ff.h"
#include "xprintf.h"
#else
#define xprintf printf
#endif

static int voomInited=0;

unsigned char *lvlData; //ptr to storage for level data

//Array to keep lines in
typedef struct {
	int x;
	int y;
} Vertex;

typedef struct {
	Vertex p[2];
} Line2d;

#define LINEMAX 150
static Line2d lines[LINEMAX];
static int lineIdx;


void veLineAdd(int xs, int ys, int xe, int ye) {
	int t;
	if (lineIdx==LINEMAX) return;
	if (xs>xe) {
		t=xs; xs=xe; xe=t;
		t=ys; ys=ye; ye=t;
	}

	if (xs<0 || xs>SIZEX) return;
	if (ys<0 || ys>SIZEX) return;
	if (xe<0 || xe>SIZEX) return;
	if (ye<0 || ye>SIZEX) return;

	if (xs==xe && ys==ye) return;

	lines[lineIdx].p[0].x=xs;
	lines[lineIdx].p[0].y=ys;
	lines[lineIdx].p[1].x=xe;
	lines[lineIdx].p[1].y=ye;
	lineIdx++;
}

static int vecAddLine(char *p, int dx, int dy, int c) {
	int sp=0x60;
	while(1) {
		if (abs(dx)>127 || abs(dy)>127) {
			sp=sp*2;
			dx=dx/2;
			dy=dy/2;
		} else if (abs(dx)<62 && abs(dy)<62 && sp>0x8) {
			sp=sp/2;
			dx=dx*2;
			dy=dy*2;
		} else {
			break;
		}
	}

	*p++=c;
#ifndef VECX
	*p++=sp-2;
#else
	*p++=sp;
#endif
	*p++=-dy;
	*p=dx;
//	printf("Added point %d,%d, spd %d int %x\n", dx, dy, sp, c);

	return 4;
}

static void linesDraw(char *cartSpace) {
	int i, j, k, d, dx, dy;
	int cd, ci, inv;
	int x=SIZEX/2,y=SIZEY/2;
	char *p=cartSpace;
	int nwr=0;

//	xprintf("Drawing %d lines.\n", lineIdx);
	for (i=0; i<lineIdx; i++) {
		nwr++;
		//Find closest line
		cd=99999; ci=0;
		for (j=0; j<lineIdx; j++) {
			if (lines[j].p[0].x!=-1) {
				for (k=0; k<2; k++) {
					dx=abs(lines[j].p[k].x-x);
					dy=abs(lines[j].p[k].y-y);
					if (dx>dy) d=dx; else d=dy;
					if (d<cd) {
						cd=d;
						ci=j;
						inv=k;
					}
				}
			}
		}

		dx=lines[ci].p[inv].x-x;
		dy=lines[ci].p[inv].y-y;
		if (dx>SIZEX/2 || dy>SIZEY/2 || nwr>3) {
			//Zero integrators.
			nwr=0;
			x=(SIZEX/2);
			y=(SIZEY/2);
			dx=lines[ci].p[inv].x-x;
			dy=lines[ci].p[inv].y-y;
		}
		if (dx!=0 || dy!=0) {
			p+=vecAddLine(p, dx, dy, (nwr==0)?0x40:0);
			x+=dx;
			y+=dy;
		}

		dx=lines[ci].p[inv^1].x-x;
		dy=lines[ci].p[inv^1].y-y;
		p+=vecAddLine(p, dx, dy, 0x3f);
		x+=dx;
		y+=dy;

		//Poison line
		lines[ci].p[0].x=-1;
	}
	*p++=0xff;
//	printf("Drew %i lines in %i bytes\n", lineIdx, p-cartSpace);
}


#ifdef VECX
static void doInitVoom() {
	FILE *f;
	char *p=lvlData;
	const char *names[]={
		"e1l1/VERTEXES",
		"e1l1/LINEDEFS",
		"e1l1/SECTORS",
		"e1l1/SSECTORS",
		"e1l1/SEGS",
		"e1l1/SIDEDEFS",
		"e1l1/NODES"
	};
	char *dptr[7];
	int dsize[7];
	int n, i, l;
	l=sizeof(lvlData);
	for (i=0; i<7; i++) {
		f=fopen(names[i], "r");
		n=fread(p, 1, l, f);
		fclose(f);
		dptr[i]=p;
		dsize[i]=n;
		p+=n;
		l=l-n;
		printf("%s: %d bytes\n", names[i], n);
	}
	printf("Total level data size: %d bytes. %d bytes left in level buffer.\n", n, l);
	voomInit(dptr, dsize);
	voomInited=1;
}
#else
static void doInitVoom() {
	char *p=lvlData;
	const char *names[]={
		"e1l1/VERTEXES",
		"e1l1/LINEDEFS",
		"e1l1/SECTORS",
		"e1l1/SSECTORS",
		"e1l1/SEGS",
		"e1l1/SIDEDEFS",
		"e1l1/NODES"
	};
	FIL f;
	FRESULT fr;
	char *dptr[7];
	int dsize[7];
	int n, i, l, t;
	l=64*1024-0x800;
	for (i=0; i<7; i++) {
		fr=f_open(&f, names[i], FA_READ);
		if (fr) return;
		f_read(&f, p, l, &n);
		xprintf("Loaded %s: %d\n", names[i], n);
		f_close(&f);
		dptr[i]=p;
		dsize[i]=n;
		p+=n;
		l=l-n;
	}

	//NOTE: This load is pretty much endian dependent and only works on any CPU
	//with the same endian-ness as the x86 Doom originally was made for.

	voomInit(dptr, dsize);
	xprintf("Loaded %d bytes.\n", n);
	voomInited=1;
}
#endif

void voomVectrexFrame(int keys, char *vecSpace, char *lvlDataPtr) {
	lvlData=lvlDataPtr;
	if (!voomInited) doInitVoom();
	voomHandleKeys(keys);
	lineIdx=0;
	voomDraw();
	linesDraw(vecSpace);
}

